# screenshot_app.tcl ---
#
# -------------------------------------------------------------------------
# (c) 2019, Johann Oberdorfer - Engineering Support | CAD | Software
#     johann.oberdorfer [at] gmail.com
#     www.johann-oberdorfer.eu
# -------------------------------------------------------------------------
# This source file is distributed under the BSD license.
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#   See the BSD License for more details.
# -------------------------------------------------------------------------
# Purpose:
#   Easy to use screenshot utility program.
#	During the day, I find myself quite often repetitly doing these steps:
#     - create a screenshot (with the "Snipping Tool"),
#	  - paste the screenshot into "Paint" and apply some visual marks, notes, etc...
#     - finally copy/paste the image to a presentation or e-mail client program.
#    So I started thinking about how I could simplify this process.
#    That was one of the reasions, why I started to develop this "screenshot_app".
#
#    You can also use the code just as an example of how to build up
#    a program in tcl/tk. The app basicly shows the steps required
#    to create a tcl/tk application which based on a few packages and some
#    custom made code.
#
#   Features:
#     - Zero installation, just one file which needs to be copied.
#     - Current version copies the screen to the clipboard for further
#       usage with an image editor (maybe Paint.exe under windows).
#     - remembers screen position and screen size
#	  - when creating the screenshot, a transparent window appears
#       which is more convinient to manipulate and adjust before the
#       sceen area is actually copied.
#     - timeout on + topmost setting off allows to screen capture e.g.
#       pull-down menus of a running application - yes (!)  
#
# -------------------------------------------------------------------------
# Release notes:
# 19-09-06, Initial release, Hans
#           Created this code in-between while commuting with the train.
#
# -------------------------------------------------------------------------
# Future development:
#
# program option menu:
#   - screen shot successfully created, what to do next:
#	    - copy image to clipoard
#	    - save image to a file
#       - open the image with Paint.exe
#
#   - add info and help functions to the gui
# -------------------------------------------------------------------------

# where to find the required library packages,
# auto_path usually needs to be modified to fit your specific environment:
#
set dir [file dirname [info script]]
lappend auto_path [file join $dir "."]
# lappend auto_path [file join $dir "../../tksqlite-0.5.13-modified.vfs/lib"]


package require Tk
package require TclOO
package require treectrl
package require Img

package require screenshot


namespace eval screenshot_app {
	variable this

	set ini_defaults {
		{"geometry" "950x570+140+170"}
		{"showgrid" "1"}
		{"opacity" "0.7"}
	}

	::savedefault::savedefault \
		"SceenShotSettings.ini" \
		$ini_defaults

	::savedefault::readsettings this

	# -unused- for the moment...
	#
	proc SaveScreenShot2File {wparent capture_img} {

		# finally, write image to file and we are done...
		set filetypes {
			{"All Image Files" {.gif .png .jpg}}
			{"PNG Images" .png}
		}

		set re {\.(gif|png)$}
		set LASTDIR [pwd]
				
		set file [tk_getSaveFile \
			-parent $wparent -title "Save Image to File" \
			-initialdir $LASTDIR -filetypes $filetypes]
				
		if {$file ne ""} {
				
			if {![regexp -nocase $re $file -> ext]} {
				set ext "png"
				append file ".${ext}"
			}
			
			if {[catch {$capture_img write $file \
						-format [string tolower $ext]} err]} {
							
				tk_messageBox -title "Error Writing File" \
					-parent $wparent -icon error -type ok \
					-message "Error writing to file \"$file\":\n$err"
			}
			
			# clear some memory:
			image delete $capture_img
		}
	}

	proc SaveScreenShot2ClipBoard { wparent \
									capture_img geometry showgrid opacity} {
		variable this

		# Use TWAPI and tkImg BMP to copy/paste image with clipboard
		# code from the tcl/tk wiki, article published by Paul Obermeier, available under:
		# https://wiki.tcl-lang.org/page/Copy+image+to+and+from+the+Windows+clipboard
		
		# First 14 bytes are bitmapfileheader - get rid of this
		set data [string range [base64::decode [$capture_img data -format bmp]] 14 end]

		twapi::open_clipboard
		twapi::empty_clipboard

		# First 14 bytes are bitmapfileheader - get rid of this
		twapi::write_clipboard 8 $data
		twapi::close_clipboard

		# make sure message appears on top, expecially important when
		# in combination with timeout and topmost off:
		catch {
			# wm withdraw $this(top)
			wm attributes $this(top) -topmost 1
			wm overrideredirect $this(top) 1
		}

		tk_messageBox -title "Screenshot Information:" \
			-parent $this(top) -icon info -type ok \
			-message "Image successfully copied to clipboard!"

		# clear some memory:
		image delete $capture_img
		
		ScreenShotCancelCmd $geometry $showgrid $opacity
	}
	
	proc ScreenShotCancelCmd {geometry showgrid opacity} {

		set ini_list {}
		lappend ini_list [list "geometry" $geometry]
		lappend ini_list [list "showgrid" $showgrid]
		lappend ini_list [list "opacity"  $opacity]

		::savedefault::savesettings $ini_list
		::cleanup_temp_for_win [::fileutil::tempdir]

		exit 0
	}
}


proc ::screenshot_app::main {} {
	variable this
	variable app_images

	wm withdraw .
	set t [toplevel .t]
	set this(top) $t
	
	bind all <F1> "[namespace current]::ShowAboutDlg"

	screenshot::screenshot $t.scrnshot \
			-background LightYellow \
			-foreground DarkGreen \
			-geometry $this(geometry) \
			-grid $this(showgrid) \
			-alpha $this(opacity) \
			-screenshotcommand "[namespace current]::SaveScreenShot2ClipBoard $t" \
			-screenshotcancelcommand "[namespace current]::ScreenShotCancelCmd"

	pack $t.scrnshot -expand true -fill both
	
	set m [$t.scrnshot getmenuwidget]
	
	$m add command \
		-label "About..." \
		-accelerator "F1" \
		-command "[namespace current]::ShowAboutDlg" \
		-image $::app_images(dialog-information) \
		-compound left
	
}

